# BSD 3-Clause License; see https://github.com/scikit-hep/awkward-1.0/blob/main/LICENSE

import numpy as np
import pytest

import awkward as ak

to_list = ak.operations.to_list


def test_listoffsetarray_localindex():
    v2_array = ak.operations.from_iter(
        [[0.0, 1.1, 2.2], [], [3.3, 4.4], [5.5], [6.6, 7.7, 8.8, 9.9]], highlevel=False
    )
    assert to_list(ak._do.local_index(v2_array, 0)) == [0, 1, 2, 3, 4]
    assert (
        ak._do.local_index(v2_array.to_typetracer(), 0).form
        == ak._do.local_index(v2_array, 0).form
    )
    assert to_list(ak._do.local_index(v2_array, 1)) == [
        [0, 1, 2],
        [],
        [0, 1],
        [0],
        [0, 1, 2, 3],
    ]
    assert (
        ak._do.local_index(v2_array.to_typetracer(), 1).form
        == ak._do.local_index(v2_array, 1).form
    )
    assert to_list(ak._do.local_index(v2_array, -1)) == [
        [0, 1, 2],
        [],
        [0, 1],
        [0],
        [0, 1, 2, 3],
    ]
    assert (
        ak._do.local_index(v2_array.to_typetracer(), -1).form
        == ak._do.local_index(v2_array, -1).form
    )
    assert to_list(ak._do.local_index(v2_array, -2)) == [0, 1, 2, 3, 4]
    assert (
        ak._do.local_index(v2_array.to_typetracer(), -2).form
        == ak._do.local_index(v2_array, -2).form
    )

    with pytest.raises(IndexError):
        ak._do.local_index(v2_array, -3)
    with pytest.raises(IndexError):
        ak._do.local_index(v2_array, 2)

    v2_array = ak.operations.from_iter(
        [[[0.0, 1.1, 2.2], [], [3.3, 4.4]], [], [[5.5]], [[6.6, 7.7, 8.8, 9.9]]],
        highlevel=False,
    )
    assert to_list(ak._do.local_index(v2_array, 0)) == [0, 1, 2, 3]
    assert (
        ak._do.local_index(v2_array.to_typetracer(), 0).form
        == ak._do.local_index(v2_array, 0).form
    )
    assert to_list(ak._do.local_index(v2_array, 1)) == [[0, 1, 2], [], [0], [0]]
    assert (
        ak._do.local_index(v2_array.to_typetracer(), 1).form
        == ak._do.local_index(v2_array, 1).form
    )
    assert to_list(ak._do.local_index(v2_array, 2)) == [
        [[0, 1, 2], [], [0, 1]],
        [],
        [[0]],
        [[0, 1, 2, 3]],
    ]
    assert (
        ak._do.local_index(v2_array.to_typetracer(), 2).form
        == ak._do.local_index(v2_array, 2).form
    )
    assert to_list(ak._do.local_index(v2_array, -1)) == [
        [[0, 1, 2], [], [0, 1]],
        [],
        [[0]],
        [[0, 1, 2, 3]],
    ]
    assert (
        ak._do.local_index(v2_array.to_typetracer(), -1).form
        == ak._do.local_index(v2_array, -1).form
    )
    assert to_list(ak._do.local_index(v2_array, -2)) == [[0, 1, 2], [], [0], [0]]
    assert (
        ak._do.local_index(v2_array.to_typetracer(), -2).form
        == ak._do.local_index(v2_array, -2).form
    )
    assert to_list(ak._do.local_index(v2_array, -3)) == [0, 1, 2, 3]
    assert (
        ak._do.local_index(v2_array.to_typetracer(), -3).form
        == ak._do.local_index(v2_array, -3).form
    )

    with pytest.raises(IndexError):
        ak._do.local_index(v2_array, -4)
    with pytest.raises(IndexError):
        ak._do.local_index(v2_array, 3)


def test_regulararray_localindex():
    v2_array = ak.operations.from_numpy(
        np.arange(2 * 3 * 5).reshape(2, 3, 5), regulararray=True, highlevel=False
    )
    assert to_list(ak._do.local_index(v2_array, 0)) == [0, 1]
    assert (
        ak._do.local_index(v2_array.to_typetracer(), 0).form
        == ak._do.local_index(v2_array, 0).form
    )
    assert to_list(ak._do.local_index(v2_array, 1)) == [[0, 1, 2], [0, 1, 2]]
    assert (
        ak._do.local_index(v2_array.to_typetracer(), 1).form
        == ak._do.local_index(v2_array, 1).form
    )
    assert to_list(ak._do.local_index(v2_array, 2)) == [
        [[0, 1, 2, 3, 4], [0, 1, 2, 3, 4], [0, 1, 2, 3, 4]],
        [[0, 1, 2, 3, 4], [0, 1, 2, 3, 4], [0, 1, 2, 3, 4]],
    ]
    assert (
        ak._do.local_index(v2_array.to_typetracer(), 2).form
        == ak._do.local_index(v2_array, 2).form
    )
    assert to_list(ak._do.local_index(v2_array, -1)) == [
        [[0, 1, 2, 3, 4], [0, 1, 2, 3, 4], [0, 1, 2, 3, 4]],
        [[0, 1, 2, 3, 4], [0, 1, 2, 3, 4], [0, 1, 2, 3, 4]],
    ]
    assert (
        ak._do.local_index(v2_array.to_typetracer(), -1).form
        == ak._do.local_index(v2_array, -1).form
    )
    assert to_list(ak._do.local_index(v2_array, -2)) == [[0, 1, 2], [0, 1, 2]]
    assert (
        ak._do.local_index(v2_array.to_typetracer(), -2).form
        == ak._do.local_index(v2_array, -2).form
    )
    assert to_list(ak._do.local_index(v2_array, -3)) == [0, 1]
    assert (
        ak._do.local_index(v2_array.to_typetracer(), -3).form
        == ak._do.local_index(v2_array, -3).form
    )

    with pytest.raises(IndexError):
        ak._do.local_index(v2_array, -4)
    with pytest.raises(IndexError):
        ak._do.local_index(v2_array, 3)

    v2_array = ak.operations.from_numpy(
        np.arange(2 * 3 * 5 * 10).reshape(2, 3, 5, 10),
        regulararray=True,
        highlevel=False,
    )
    assert to_list(ak._do.local_index(v2_array, 0)) == [0, 1]
    assert (
        ak._do.local_index(v2_array.to_typetracer(), 0).form
        == ak._do.local_index(v2_array, 0).form
    )
    assert to_list(ak._do.local_index(v2_array, 1)) == [[0, 1, 2], [0, 1, 2]]
    assert (
        ak._do.local_index(v2_array.to_typetracer(), 1).form
        == ak._do.local_index(v2_array, 1).form
    )
    assert to_list(ak._do.local_index(v2_array, 2)) == [
        [[0, 1, 2, 3, 4], [0, 1, 2, 3, 4], [0, 1, 2, 3, 4]],
        [[0, 1, 2, 3, 4], [0, 1, 2, 3, 4], [0, 1, 2, 3, 4]],
    ]
    assert (
        ak._do.local_index(v2_array.to_typetracer(), 2).form
        == ak._do.local_index(v2_array, 2).form
    )
    assert to_list(ak._do.local_index(v2_array, 3)) == [
        [
            [
                [0, 1, 2, 3, 4, 5, 6, 7, 8, 9],
                [0, 1, 2, 3, 4, 5, 6, 7, 8, 9],
                [0, 1, 2, 3, 4, 5, 6, 7, 8, 9],
                [0, 1, 2, 3, 4, 5, 6, 7, 8, 9],
                [0, 1, 2, 3, 4, 5, 6, 7, 8, 9],
            ],
            [
                [0, 1, 2, 3, 4, 5, 6, 7, 8, 9],
                [0, 1, 2, 3, 4, 5, 6, 7, 8, 9],
                [0, 1, 2, 3, 4, 5, 6, 7, 8, 9],
                [0, 1, 2, 3, 4, 5, 6, 7, 8, 9],
                [0, 1, 2, 3, 4, 5, 6, 7, 8, 9],
            ],
            [
                [0, 1, 2, 3, 4, 5, 6, 7, 8, 9],
                [0, 1, 2, 3, 4, 5, 6, 7, 8, 9],
                [0, 1, 2, 3, 4, 5, 6, 7, 8, 9],
                [0, 1, 2, 3, 4, 5, 6, 7, 8, 9],
                [0, 1, 2, 3, 4, 5, 6, 7, 8, 9],
            ],
        ],
        [
            [
                [0, 1, 2, 3, 4, 5, 6, 7, 8, 9],
                [0, 1, 2, 3, 4, 5, 6, 7, 8, 9],
                [0, 1, 2, 3, 4, 5, 6, 7, 8, 9],
                [0, 1, 2, 3, 4, 5, 6, 7, 8, 9],
                [0, 1, 2, 3, 4, 5, 6, 7, 8, 9],
            ],
            [
                [0, 1, 2, 3, 4, 5, 6, 7, 8, 9],
                [0, 1, 2, 3, 4, 5, 6, 7, 8, 9],
                [0, 1, 2, 3, 4, 5, 6, 7, 8, 9],
                [0, 1, 2, 3, 4, 5, 6, 7, 8, 9],
                [0, 1, 2, 3, 4, 5, 6, 7, 8, 9],
            ],
            [
                [0, 1, 2, 3, 4, 5, 6, 7, 8, 9],
                [0, 1, 2, 3, 4, 5, 6, 7, 8, 9],
                [0, 1, 2, 3, 4, 5, 6, 7, 8, 9],
                [0, 1, 2, 3, 4, 5, 6, 7, 8, 9],
                [0, 1, 2, 3, 4, 5, 6, 7, 8, 9],
            ],
        ],
    ]
    assert (
        ak._do.local_index(v2_array.to_typetracer(), 3).form
        == ak._do.local_index(v2_array, 3).form
    )
    assert to_list(ak._do.local_index(v2_array, -1)) == [
        [
            [
                [0, 1, 2, 3, 4, 5, 6, 7, 8, 9],
                [0, 1, 2, 3, 4, 5, 6, 7, 8, 9],
                [0, 1, 2, 3, 4, 5, 6, 7, 8, 9],
                [0, 1, 2, 3, 4, 5, 6, 7, 8, 9],
                [0, 1, 2, 3, 4, 5, 6, 7, 8, 9],
            ],
            [
                [0, 1, 2, 3, 4, 5, 6, 7, 8, 9],
                [0, 1, 2, 3, 4, 5, 6, 7, 8, 9],
                [0, 1, 2, 3, 4, 5, 6, 7, 8, 9],
                [0, 1, 2, 3, 4, 5, 6, 7, 8, 9],
                [0, 1, 2, 3, 4, 5, 6, 7, 8, 9],
            ],
            [
                [0, 1, 2, 3, 4, 5, 6, 7, 8, 9],
                [0, 1, 2, 3, 4, 5, 6, 7, 8, 9],
                [0, 1, 2, 3, 4, 5, 6, 7, 8, 9],
                [0, 1, 2, 3, 4, 5, 6, 7, 8, 9],
                [0, 1, 2, 3, 4, 5, 6, 7, 8, 9],
            ],
        ],
        [
            [
                [0, 1, 2, 3, 4, 5, 6, 7, 8, 9],
                [0, 1, 2, 3, 4, 5, 6, 7, 8, 9],
                [0, 1, 2, 3, 4, 5, 6, 7, 8, 9],
                [0, 1, 2, 3, 4, 5, 6, 7, 8, 9],
                [0, 1, 2, 3, 4, 5, 6, 7, 8, 9],
            ],
            [
                [0, 1, 2, 3, 4, 5, 6, 7, 8, 9],
                [0, 1, 2, 3, 4, 5, 6, 7, 8, 9],
                [0, 1, 2, 3, 4, 5, 6, 7, 8, 9],
                [0, 1, 2, 3, 4, 5, 6, 7, 8, 9],
                [0, 1, 2, 3, 4, 5, 6, 7, 8, 9],
            ],
            [
                [0, 1, 2, 3, 4, 5, 6, 7, 8, 9],
                [0, 1, 2, 3, 4, 5, 6, 7, 8, 9],
                [0, 1, 2, 3, 4, 5, 6, 7, 8, 9],
                [0, 1, 2, 3, 4, 5, 6, 7, 8, 9],
                [0, 1, 2, 3, 4, 5, 6, 7, 8, 9],
            ],
        ],
    ]
    assert (
        ak._do.local_index(v2_array.to_typetracer(), -1).form
        == ak._do.local_index(v2_array, -1).form
    )
    assert to_list(ak._do.local_index(v2_array, -2)) == [
        [[0, 1, 2, 3, 4], [0, 1, 2, 3, 4], [0, 1, 2, 3, 4]],
        [[0, 1, 2, 3, 4], [0, 1, 2, 3, 4], [0, 1, 2, 3, 4]],
    ]
    assert (
        ak._do.local_index(v2_array.to_typetracer(), -2).form
        == ak._do.local_index(v2_array, -2).form
    )
    assert to_list(ak._do.local_index(v2_array, -3)) == [[0, 1, 2], [0, 1, 2]]
    assert (
        ak._do.local_index(v2_array.to_typetracer(), -3).form
        == ak._do.local_index(v2_array, -3).form
    )
    assert to_list(ak._do.local_index(v2_array, -4)) == [0, 1]
    assert (
        ak._do.local_index(v2_array.to_typetracer(), -4).form
        == ak._do.local_index(v2_array, -4).form
    )

    with pytest.raises(IndexError):
        ak._do.local_index(v2_array, -5)
    with pytest.raises(IndexError):
        ak._do.local_index(v2_array, 4)

    v2_array = ak.highlevel.Array(
        ak.contents.RegularArray(
            ak.highlevel.Array([[1, 2, 3], [], [4, 5]]).layout, 0, zeros_length=0
        )
    ).layout

    assert to_list(ak._do.local_index(v2_array, 0)) == []
    assert (
        ak._do.local_index(v2_array.to_typetracer(), 0).form
        == ak._do.local_index(v2_array, 0).form
    )
    assert to_list(ak._do.local_index(v2_array, 1)) == []
    assert (
        ak._do.local_index(v2_array.to_typetracer(), 1).form
        == ak._do.local_index(v2_array, 1).form
    )
    assert to_list(ak._do.local_index(v2_array, 2)) == []
    assert (
        ak._do.local_index(v2_array.to_typetracer(), 2).form
        == ak._do.local_index(v2_array, 2).form
    )
    assert to_list(ak._do.local_index(v2_array, -1)) == []
    assert (
        ak._do.local_index(v2_array.to_typetracer(), -1).form
        == ak._do.local_index(v2_array, -1).form
    )
    assert to_list(ak._do.local_index(v2_array, -2)) == []
    assert (
        ak._do.local_index(v2_array.to_typetracer(), -2).form
        == ak._do.local_index(v2_array, -2).form
    )
    assert to_list(ak._do.local_index(v2_array, -3)) == []
    assert (
        ak._do.local_index(v2_array.to_typetracer(), -3).form
        == ak._do.local_index(v2_array, -3).form
    )

    with pytest.raises(IndexError):
        ak._do.local_index(v2_array, -4)
    with pytest.raises(IndexError):
        ak._do.local_index(v2_array, 3)


def test_bytemaskedarray_localindex():
    content = ak.operations.from_iter(
        [
            [[0.0, 1.1, 2.2], [], [3.3, 4.4]],
            [],
            [[5.5]],
            [[6.6, 7.7, 8.8, 9.9]],
            [[], [10.0, 11.1, 12.2]],
        ],
        highlevel=False,
    )
    mask = ak.index.Index8(np.array([0, 0, 1, 1, 0], dtype=np.int8))
    v2_array = ak.contents.ByteMaskedArray(mask, content, valid_when=False)

    assert to_list(v2_array) == [
        [[0.0, 1.1, 2.2], [], [3.3, 4.4]],
        [],
        None,
        None,
        [[], [10.0, 11.1, 12.2]],
    ]
    assert to_list(ak._do.local_index(v2_array, axis=0)) == [0, 1, 2, 3, 4]
    assert (
        ak._do.local_index(v2_array.to_typetracer(), axis=0).form
        == ak._do.local_index(v2_array, axis=0).form
    )
    assert to_list(ak._do.local_index(v2_array, axis=-3)) == [0, 1, 2, 3, 4]
    assert (
        ak._do.local_index(v2_array.to_typetracer(), axis=-3).form
        == ak._do.local_index(v2_array, axis=-3).form
    )
    assert to_list(ak._do.local_index(v2_array, axis=1)) == [
        [0, 1, 2],
        [],
        None,
        None,
        [0, 1],
    ]
    assert (
        ak._do.local_index(v2_array.to_typetracer(), axis=1).form
        == ak._do.local_index(v2_array, axis=1).form
    )
    assert to_list(ak._do.local_index(v2_array, axis=-2)) == [
        [0, 1, 2],
        [],
        None,
        None,
        [0, 1],
    ]
    assert (
        ak._do.local_index(v2_array.to_typetracer(), axis=-2).form
        == ak._do.local_index(v2_array, axis=-2).form
    )
    assert to_list(ak._do.local_index(v2_array, axis=2)) == [
        [[0, 1, 2], [], [0, 1]],
        [],
        None,
        None,
        [[], [0, 1, 2]],
    ]
    assert (
        ak._do.local_index(v2_array.to_typetracer(), axis=2).form
        == ak._do.local_index(v2_array, axis=2).form
    )
    assert to_list(ak._do.local_index(v2_array, axis=-1)) == [
        [[0, 1, 2], [], [0, 1]],
        [],
        None,
        None,
        [[], [0, 1, 2]],
    ]
    assert (
        ak._do.local_index(v2_array.to_typetracer(), axis=-1).form
        == ak._do.local_index(v2_array, axis=-1).form
    )

    with pytest.raises(IndexError):
        ak._do.local_index(v2_array, axis=4)
    with pytest.raises(IndexError):
        ak._do.local_index(v2_array, axis=-4)


def test_numpyarray_localindex():
    v2_array = ak.contents.numpyarray.NumpyArray(
        np.array([0.0, 1.1, 2.2, 3.3], dtype=np.float64)
    )
    assert to_list(ak._do.local_index(v2_array, axis=0)) == [0, 1, 2, 3]
    assert (
        ak._do.local_index(v2_array.to_typetracer(), axis=0).form
        == ak._do.local_index(v2_array, axis=0).form
    )
    assert to_list(ak._do.local_index(v2_array, axis=-1)) == [0, 1, 2, 3]
    assert (
        ak._do.local_index(v2_array.to_typetracer(), axis=-1).form
        == ak._do.local_index(v2_array, axis=-1).form
    )

    with pytest.raises(IndexError):
        ak._do.local_index(v2_array, axis=1)
    with pytest.raises(IndexError):
        ak._do.local_index(v2_array, axis=2)
    with pytest.raises(IndexError):
        ak._do.local_index(v2_array, axis=-2)


def test_bitmaskedarray_localindex():
    v2_array = ak.contents.bitmaskedarray.BitMaskedArray(
        ak.index.Index(
            np.packbits(
                np.array(
                    [
                        1,
                        1,
                        1,
                        1,
                        0,
                        0,
                        0,
                        0,
                        1,
                        0,
                        1,
                        0,
                        1,
                    ],
                    dtype=np.uint8,
                )
            )
        ),
        ak.contents.numpyarray.NumpyArray(
            np.array(
                [0.0, 1.0, 2.0, 3.0, 4.0, 5.0, 6.0, 7.0, 1.1, 2.2, 3.3, 4.4, 5.5, 6.6]
            )
        ),
        valid_when=True,
        length=13,
        lsb_order=False,
    )

    assert to_list(ak._do.local_index(v2_array, 0)) == [
        0,
        1,
        2,
        3,
        4,
        5,
        6,
        7,
        8,
        9,
        10,
        11,
        12,
    ]
    assert to_list(ak._do.local_index(v2_array, -1)) == [
        0,
        1,
        2,
        3,
        4,
        5,
        6,
        7,
        8,
        9,
        10,
        11,
        12,
    ]
    assert (
        ak._do.local_index(v2_array.to_typetracer(), -1).form
        == ak._do.local_index(v2_array, -1).form
    )

    with pytest.raises(IndexError):
        ak._do.local_index(v2_array, -2)
    with pytest.raises(IndexError):
        ak._do.local_index(v2_array, 1)

    v2_array = ak.contents.bitmaskedarray.BitMaskedArray(
        ak.index.Index(
            np.packbits(
                np.array(
                    [
                        0,
                        0,
                        0,
                        0,
                        1,
                        1,
                        1,
                        1,
                        0,
                        1,
                        0,
                        1,
                        0,
                    ],
                    dtype=np.uint8,
                )
            )
        ),
        ak.contents.numpyarray.NumpyArray(
            np.array(
                [0.0, 1.0, 2.0, 3.0, 4.0, 5.0, 6.0, 7.0, 1.1, 2.2, 3.3, 4.4, 5.5, 6.6]
            )
        ),
        valid_when=False,
        length=13,
        lsb_order=False,
    )

    assert to_list(ak._do.local_index(v2_array, 0)) == [
        0,
        1,
        2,
        3,
        4,
        5,
        6,
        7,
        8,
        9,
        10,
        11,
        12,
    ]
    assert (
        ak._do.local_index(v2_array.to_typetracer(), 0).form
        == ak._do.local_index(v2_array, 0).form
    )
    assert to_list(ak._do.local_index(v2_array, -1)) == [
        0,
        1,
        2,
        3,
        4,
        5,
        6,
        7,
        8,
        9,
        10,
        11,
        12,
    ]
    assert (
        ak._do.local_index(v2_array.to_typetracer(), -1).form
        == ak._do.local_index(v2_array, -1).form
    )

    with pytest.raises(IndexError):
        ak._do.local_index(v2_array, -2)
    with pytest.raises(IndexError):
        ak._do.local_index(v2_array, 1)

    v2_array = ak.contents.bitmaskedarray.BitMaskedArray(
        ak.index.Index(
            np.packbits(
                np.array(
                    [
                        0,
                        0,
                        0,
                        0,
                        1,
                        1,
                        1,
                        1,
                        0,
                        0,
                        0,
                        1,
                        0,
                        1,
                        0,
                        1,
                    ],
                    dtype=np.uint8,
                )
            )
        ),
        ak.contents.numpyarray.NumpyArray(
            np.array(
                [0.0, 1.0, 2.0, 3.0, 4.0, 5.0, 6.0, 7.0, 1.1, 2.2, 3.3, 4.4, 5.5, 6.6]
            )
        ),
        valid_when=True,
        length=13,
        lsb_order=True,
    )
    assert to_list(ak._do.local_index(v2_array, 0)) == [
        0,
        1,
        2,
        3,
        4,
        5,
        6,
        7,
        8,
        9,
        10,
        11,
        12,
    ]
    assert (
        ak._do.local_index(v2_array.to_typetracer(), 0).form
        == ak._do.local_index(v2_array, 0).form
    )
    assert to_list(ak._do.local_index(v2_array, -1)) == [
        0,
        1,
        2,
        3,
        4,
        5,
        6,
        7,
        8,
        9,
        10,
        11,
        12,
    ]
    assert (
        ak._do.local_index(v2_array.to_typetracer(), -1).form
        == ak._do.local_index(v2_array, -1).form
    )

    with pytest.raises(IndexError):
        ak._do.local_index(v2_array, -2)
    with pytest.raises(IndexError):
        ak._do.local_index(v2_array, 1)


def test_unmaskedarray_localindex():
    v2_array = ak.contents.unmaskedarray.UnmaskedArray(
        ak.contents.numpyarray.NumpyArray(
            np.array([0.0, 1.1, 2.2, 3.3], dtype=np.float64)
        )
    )
    assert to_list(ak._do.local_index(v2_array, 0)) == [0, 1, 2, 3]
    assert (
        ak._do.local_index(v2_array.to_typetracer(), 0).form
        == ak._do.local_index(v2_array, 0).form
    )
    assert to_list(ak._do.local_index(v2_array, -1)) == [0, 1, 2, 3]
    assert (
        ak._do.local_index(v2_array.to_typetracer(), -1).form
        == ak._do.local_index(v2_array, -1).form
    )

    with pytest.raises(IndexError):
        ak._do.local_index(v2_array, -2)
    with pytest.raises(IndexError):
        ak._do.local_index(v2_array, 1)


def test_unionarray_localindex():
    v2_array = ak.contents.unionarray.UnionArray(
        ak.index.Index(np.array([1, 1, 0, 0, 1, 0, 1], dtype=np.int8)),
        ak.index.Index(np.array([4, 3, 0, 1, 2, 2, 4, 100])),
        [
            ak.from_iter(["1", "2", "3"], highlevel=False),
            ak.contents.numpyarray.NumpyArray(np.array([1.1, 2.2, 3.3, 4.4, 5.5])),
        ],
    )
    assert to_list(ak._do.local_index(v2_array, 0)) == [0, 1, 2, 3, 4, 5, 6]
    assert (
        ak._do.local_index(v2_array.to_typetracer(), 0).form
        == ak._do.local_index(v2_array, 0).form
    )
    assert to_list(ak._do.local_index(v2_array, -1)) == [0, 1, 2, 3, 4, 5, 6]
    assert (
        ak._do.local_index(v2_array.to_typetracer(), -1).form
        == ak._do.local_index(v2_array, -1).form
    )

    with pytest.raises(IndexError):
        ak._do.local_index(v2_array, -2)
    with pytest.raises(IndexError):
        ak._do.local_index(v2_array, 1)


def test_recordarray_localindex():
    v2_array = ak.contents.regulararray.RegularArray(
        ak.contents.recordarray.RecordArray(
            [
                ak.contents.numpyarray.NumpyArray(
                    np.array([0.0, 1.1, 2.2, 3.3, 4.4, 5.5, 6.6])
                )
            ],
            ["nest"],
        ),
        3,
    )
    assert to_list(ak._do.local_index(v2_array, 0)) == [0, 1]
    assert (
        ak._do.local_index(v2_array.to_typetracer(), 0).form
        == ak._do.local_index(v2_array, 0).form
    )
    assert to_list(ak._do.local_index(v2_array, 1)) == [[0, 1, 2], [0, 1, 2]]
    assert (
        ak._do.local_index(v2_array.to_typetracer(), 1).form
        == ak._do.local_index(v2_array, 1).form
    )
    assert to_list(ak._do.local_index(v2_array, -1)) == [[0, 1, 2], [0, 1, 2]]
    assert (
        ak._do.local_index(v2_array.to_typetracer(), -1).form
        == ak._do.local_index(v2_array, -1).form
    )
    assert to_list(ak._do.local_index(v2_array, -2)) == [0, 1]
    assert (
        ak._do.local_index(v2_array.to_typetracer(), -2).form
        == ak._do.local_index(v2_array, -2).form
    )

    with pytest.raises(IndexError):
        ak._do.local_index(v2_array, -3)
    with pytest.raises(IndexError):
        ak._do.local_index(v2_array, 2)

    v2_array = ak.contents.regulararray.RegularArray(
        ak.contents.recordarray.RecordArray(
            [ak.contents.emptyarray.EmptyArray()], ["nest"]
        ),
        0,
        zeros_length=10,
    )

    assert to_list(ak._do.local_index(v2_array, 0)) == [
        0,
        1,
        2,
        3,
        4,
        5,
        6,
        7,
        8,
        9,
    ]
    assert (
        ak._do.local_index(v2_array.to_typetracer(), 0).form
        == ak._do.local_index(v2_array, 0).form
    )
    assert to_list(ak._do.local_index(v2_array, 1)) == [
        [],
        [],
        [],
        [],
        [],
        [],
        [],
        [],
        [],
        [],
    ]
    assert (
        ak._do.local_index(v2_array.to_typetracer(), 1).form
        == ak._do.local_index(v2_array, 1).form
    )
    assert to_list(ak._do.local_index(v2_array, -1)) == [
        [],
        [],
        [],
        [],
        [],
        [],
        [],
        [],
        [],
        [],
    ]
    assert (
        ak._do.local_index(v2_array.to_typetracer(), -1).form
        == ak._do.local_index(v2_array, -1).form
    )
    assert to_list(ak._do.local_index(v2_array, -2)) == [
        0,
        1,
        2,
        3,
        4,
        5,
        6,
        7,
        8,
        9,
    ]
    assert (
        ak._do.local_index(v2_array.to_typetracer(), -2).form
        == ak._do.local_index(v2_array, -2).form
    )

    with pytest.raises(IndexError):
        ak._do.local_index(v2_array, -3)

    v2_array = ak.contents.listarray.ListArray(
        ak.index.Index(np.array([4, 100, 1])),
        ak.index.Index(np.array([7, 100, 3, 200])),
        ak.contents.recordarray.RecordArray(
            [
                ak.contents.numpyarray.NumpyArray(
                    np.array([6.6, 4.4, 5.5, 7.7, 1.1, 2.2, 3.3, 8.8])
                )
            ],
            ["nest"],
        ),
    )

    assert to_list(ak._do.local_index(v2_array, 0)) == [0, 1, 2]
    assert (
        ak._do.local_index(v2_array.to_typetracer(), 0).form
        == ak._do.local_index(v2_array, 0).form
    )
    assert to_list(ak._do.local_index(v2_array, 1)) == [[0, 1, 2], [], [0, 1]]
    assert (
        ak._do.local_index(v2_array.to_typetracer(), 1).form
        == ak._do.local_index(v2_array, 1).form
    )
    assert to_list(ak._do.local_index(v2_array, -1)) == [[0, 1, 2], [], [0, 1]]
    assert (
        ak._do.local_index(v2_array.to_typetracer(), -1).form
        == ak._do.local_index(v2_array, -1).form
    )
    assert to_list(ak._do.local_index(v2_array, -2)) == [0, 1, 2]
    assert (
        ak._do.local_index(v2_array.to_typetracer(), -2).form
        == ak._do.local_index(v2_array, -2).form
    )

    with pytest.raises(IndexError):
        ak._do.local_index(v2_array, -3)
    with pytest.raises(IndexError):
        ak._do.local_index(v2_array, 2)

    v2_array = ak.contents.listoffsetarray.ListOffsetArray(
        ak.index.Index(np.array([1, 4, 4, 6])),
        ak.contents.recordarray.RecordArray(
            [
                ak.contents.numpyarray.NumpyArray(
                    np.array([6.6, 1.1, 2.2, 3.3, 4.4, 5.5, 7.7])
                )
            ],
            ["nest"],
        ),
    )
    assert to_list(ak._do.local_index(v2_array, 0)) == [0, 1, 2]
    assert (
        ak._do.local_index(v2_array.to_typetracer(), 0).form
        == ak._do.local_index(v2_array, 0).form
    )
    assert to_list(ak._do.local_index(v2_array, 1)) == [[0, 1, 2], [], [0, 1]]
    assert (
        ak._do.local_index(v2_array.to_typetracer(), 1).form
        == ak._do.local_index(v2_array, 1).form
    )
    assert to_list(ak._do.local_index(v2_array, -1)) == [[0, 1, 2], [], [0, 1]]
    assert (
        ak._do.local_index(v2_array.to_typetracer(), -1).form
        == ak._do.local_index(v2_array, -1).form
    )
    assert to_list(ak._do.local_index(v2_array, -2)) == [0, 1, 2]
    assert (
        ak._do.local_index(v2_array.to_typetracer(), -2).form
        == ak._do.local_index(v2_array, -2).form
    )

    with pytest.raises(IndexError):
        ak._do.local_index(v2_array, -3)
    with pytest.raises(IndexError):
        ak._do.local_index(v2_array, 2)

    v2_array = ak.contents.indexedarray.IndexedArray(
        ak.index.Index(np.array([2, 2, 0, 1, 4, 5, 4])),
        ak.contents.recordarray.RecordArray(
            [
                ak.contents.numpyarray.NumpyArray(
                    np.array([1.1, 2.2, 3.3, 4.4, 5.5, 6.6])
                )
            ],
            ["nest"],
        ),
    )
    assert to_list(ak._do.local_index(v2_array, 0)) == [0, 1, 2, 3, 4, 5, 6]
    assert (
        ak._do.local_index(v2_array.to_typetracer(), 0).form
        == ak._do.local_index(v2_array, 0).form
    )
    assert to_list(ak._do.local_index(v2_array, -1)) == [0, 1, 2, 3, 4, 5, 6]
    assert (
        ak._do.local_index(v2_array.to_typetracer(), -1).form
        == ak._do.local_index(v2_array, -1).form
    )

    with pytest.raises(IndexError):
        ak._do.local_index(v2_array, -2)
    with pytest.raises(IndexError):
        ak._do.local_index(v2_array, 1)

    v2_array = ak.contents.indexedoptionarray.IndexedOptionArray(
        ak.index.Index(np.array([2, 2, -1, 1, -1, 5, 4])),
        ak.contents.recordarray.RecordArray(
            [
                ak.contents.numpyarray.NumpyArray(
                    np.array([1.1, 2.2, 3.3, 4.4, 5.5, 6.6])
                )
            ],
            ["nest"],
        ),
    )
    assert to_list(ak._do.local_index(v2_array, 0)) == [0, 1, 2, 3, 4, 5, 6]
    assert (
        ak._do.local_index(v2_array.to_typetracer(), 0).form
        == ak._do.local_index(v2_array, 0).form
    )
    assert to_list(ak._do.local_index(v2_array, -1)) == [0, 1, 2, 3, 4, 5, 6]
    assert (
        ak._do.local_index(v2_array.to_typetracer(), -1).form
        == ak._do.local_index(v2_array, -1).form
    )

    with pytest.raises(IndexError):
        ak._do.local_index(v2_array, -2)
    with pytest.raises(IndexError):
        ak._do.local_index(v2_array, 1)

    v2_array = ak.contents.bytemaskedarray.ByteMaskedArray(
        ak.index.Index(np.array([1, 0, 1, 0, 1], dtype=np.int8)),
        ak.contents.recordarray.RecordArray(
            [
                ak.contents.numpyarray.NumpyArray(
                    np.array([1.1, 2.2, 3.3, 4.4, 5.5, 6.6])
                )
            ],
            ["nest"],
        ),
        valid_when=True,
    )

    assert to_list(ak._do.local_index(v2_array, 0)) == [0, 1, 2, 3, 4]
    assert (
        ak._do.local_index(v2_array.to_typetracer(), 0).form
        == ak._do.local_index(v2_array, 0).form
    )
    assert to_list(ak._do.local_index(v2_array, -1)) == [0, 1, 2, 3, 4]
    assert (
        ak._do.local_index(v2_array.to_typetracer(), -1).form
        == ak._do.local_index(v2_array, -1).form
    )

    with pytest.raises(IndexError):
        ak._do.local_index(v2_array, -2)
    with pytest.raises(IndexError):
        ak._do.local_index(v2_array, 1)

    v2_array = ak.contents.bytemaskedarray.ByteMaskedArray(
        ak.index.Index(np.array([0, 1, 0, 1, 0], dtype=np.int8)),
        ak.contents.recordarray.RecordArray(
            [
                ak.contents.numpyarray.NumpyArray(
                    np.array([1.1, 2.2, 3.3, 4.4, 5.5, 6.6])
                )
            ],
            ["nest"],
        ),
        valid_when=False,
    )

    assert to_list(ak._do.local_index(v2_array, 0)) == [0, 1, 2, 3, 4]
    assert (
        ak._do.local_index(v2_array.to_typetracer(), 0).form
        == ak._do.local_index(v2_array, 0).form
    )
    assert to_list(ak._do.local_index(v2_array, -1)) == [0, 1, 2, 3, 4]
    assert (
        ak._do.local_index(v2_array.to_typetracer(), -1).form
        == ak._do.local_index(v2_array, -1).form
    )

    with pytest.raises(IndexError):
        ak._do.local_index(v2_array, -2)
    with pytest.raises(IndexError):
        ak._do.local_index(v2_array, 1)

    v2_array = ak.contents.bitmaskedarray.BitMaskedArray(
        ak.index.Index(
            np.packbits(
                np.array(
                    [
                        True,
                        True,
                        True,
                        True,
                        False,
                        False,
                        False,
                        False,
                        True,
                        False,
                        True,
                        False,
                        True,
                    ]
                )
            )
        ),
        ak.contents.recordarray.RecordArray(
            [
                ak.contents.numpyarray.NumpyArray(
                    np.array(
                        [
                            0.0,
                            1.0,
                            2.0,
                            3.0,
                            4.0,
                            5.0,
                            6.0,
                            7.0,
                            1.1,
                            2.2,
                            3.3,
                            4.4,
                            5.5,
                            6.6,
                        ]
                    )
                )
            ],
            ["nest"],
        ),
        valid_when=True,
        length=13,
        lsb_order=False,
    )

    assert to_list(ak._do.local_index(v2_array, 0)) == [
        0,
        1,
        2,
        3,
        4,
        5,
        6,
        7,
        8,
        9,
        10,
        11,
        12,
    ]
    assert (
        ak._do.local_index(v2_array.to_typetracer(), 0).form
        == ak._do.local_index(v2_array, 0).form
    )
    assert to_list(ak._do.local_index(v2_array, -1)) == [
        0,
        1,
        2,
        3,
        4,
        5,
        6,
        7,
        8,
        9,
        10,
        11,
        12,
    ]
    assert (
        ak._do.local_index(v2_array.to_typetracer(), -1).form
        == ak._do.local_index(v2_array, -1).form
    )
    with pytest.raises(IndexError):
        ak._do.local_index(v2_array, -2)
    with pytest.raises(IndexError):
        ak._do.local_index(v2_array, 1)

    v2_array = ak.contents.bitmaskedarray.BitMaskedArray(
        ak.index.Index(
            np.packbits(
                np.array(
                    [
                        0,
                        0,
                        0,
                        0,
                        1,
                        1,
                        1,
                        1,
                        0,
                        1,
                        0,
                        1,
                        0,
                    ],
                    dtype=np.uint8,
                )
            )
        ),
        ak.contents.recordarray.RecordArray(
            [
                ak.contents.numpyarray.NumpyArray(
                    np.array(
                        [
                            0.0,
                            1.0,
                            2.0,
                            3.0,
                            4.0,
                            5.0,
                            6.0,
                            7.0,
                            1.1,
                            2.2,
                            3.3,
                            4.4,
                            5.5,
                            6.6,
                        ]
                    )
                )
            ],
            ["nest"],
        ),
        valid_when=False,
        length=13,
        lsb_order=False,
    )

    assert to_list(ak._do.local_index(v2_array, 0)) == [
        0,
        1,
        2,
        3,
        4,
        5,
        6,
        7,
        8,
        9,
        10,
        11,
        12,
    ]
    assert (
        ak._do.local_index(v2_array.to_typetracer(), 0).form
        == ak._do.local_index(v2_array, 0).form
    )
    assert to_list(ak._do.local_index(v2_array, -1)) == [
        0,
        1,
        2,
        3,
        4,
        5,
        6,
        7,
        8,
        9,
        10,
        11,
        12,
    ]
    assert (
        ak._do.local_index(v2_array.to_typetracer(), -1).form
        == ak._do.local_index(v2_array, -1).form
    )

    with pytest.raises(IndexError):
        ak._do.local_index(v2_array, -2)
    with pytest.raises(IndexError):
        ak._do.local_index(v2_array, 1)

    v2_array = ak.contents.bitmaskedarray.BitMaskedArray(
        ak.index.Index(
            np.packbits(
                np.array(
                    [
                        0,
                        0,
                        0,
                        0,
                        1,
                        1,
                        1,
                        1,
                        0,
                        0,
                        0,
                        1,
                        0,
                        1,
                        0,
                        1,
                    ],
                    dtype=np.uint8,
                )
            )
        ),
        ak.contents.recordarray.RecordArray(
            [
                ak.contents.numpyarray.NumpyArray(
                    np.array(
                        [
                            0.0,
                            1.0,
                            2.0,
                            3.0,
                            4.0,
                            5.0,
                            6.0,
                            7.0,
                            1.1,
                            2.2,
                            3.3,
                            4.4,
                            5.5,
                            6.6,
                        ]
                    )
                )
            ],
            ["nest"],
        ),
        valid_when=True,
        length=13,
        lsb_order=True,
    )

    assert to_list(ak._do.local_index(v2_array, 0)) == [
        0,
        1,
        2,
        3,
        4,
        5,
        6,
        7,
        8,
        9,
        10,
        11,
        12,
    ]
    assert (
        ak._do.local_index(v2_array.to_typetracer(), 0).form
        == ak._do.local_index(v2_array, 0).form
    )
    assert to_list(ak._do.local_index(v2_array, -1)) == [
        0,
        1,
        2,
        3,
        4,
        5,
        6,
        7,
        8,
        9,
        10,
        11,
        12,
    ]
    assert (
        ak._do.local_index(v2_array.to_typetracer(), -1).form
        == ak._do.local_index(v2_array, -1).form
    )

    with pytest.raises(IndexError):
        ak._do.local_index(v2_array, -2)
    with pytest.raises(IndexError):
        ak._do.local_index(v2_array, 1)

    v2_array = ak.contents.bitmaskedarray.BitMaskedArray(
        ak.index.Index(
            np.packbits(
                np.array(
                    [
                        1,
                        1,
                        1,
                        1,
                        0,
                        0,
                        0,
                        0,
                        0,
                        0,
                        0,
                        0,
                        1,
                        0,
                        1,
                        0,
                    ],
                    dtype=np.uint8,
                )
            )
        ),
        ak.contents.recordarray.RecordArray(
            [
                ak.contents.numpyarray.NumpyArray(
                    np.array(
                        [
                            0.0,
                            1.0,
                            2.0,
                            3.0,
                            4.0,
                            5.0,
                            6.0,
                            7.0,
                            1.1,
                            2.2,
                            3.3,
                            4.4,
                            5.5,
                            6.6,
                        ]
                    )
                )
            ],
            ["nest"],
        ),
        valid_when=False,
        length=13,
        lsb_order=True,
    )

    assert to_list(ak._do.local_index(v2_array, 0)) == [
        0,
        1,
        2,
        3,
        4,
        5,
        6,
        7,
        8,
        9,
        10,
        11,
        12,
    ]
    assert (
        ak._do.local_index(v2_array.to_typetracer(), 0).form
        == ak._do.local_index(v2_array, 0).form
    )
    assert to_list(ak._do.local_index(v2_array, -1)) == [
        0,
        1,
        2,
        3,
        4,
        5,
        6,
        7,
        8,
        9,
        10,
        11,
        12,
    ]
    assert (
        ak._do.local_index(v2_array.to_typetracer(), -1).form
        == ak._do.local_index(v2_array, -1).form
    )

    with pytest.raises(IndexError):
        ak._do.local_index(v2_array, -2)
    with pytest.raises(IndexError):
        ak._do.local_index(v2_array, 1)

    v2_array = ak.contents.unmaskedarray.UnmaskedArray(
        ak.contents.recordarray.RecordArray(
            [
                ak.contents.numpyarray.NumpyArray(
                    np.array([0.0, 1.1, 2.2, 3.3], dtype=np.float64)
                )
            ],
            ["nest"],
        )
    )

    assert to_list(ak._do.local_index(v2_array, 0)) == [0, 1, 2, 3]
    assert (
        ak._do.local_index(v2_array.to_typetracer(), 0).form
        == ak._do.local_index(v2_array, 0).form
    )
    assert to_list(ak._do.local_index(v2_array, -1)) == [0, 1, 2, 3]
    assert (
        ak._do.local_index(v2_array.to_typetracer(), -1).form
        == ak._do.local_index(v2_array, -1).form
    )

    with pytest.raises(IndexError):
        ak._do.local_index(v2_array, -2)
    with pytest.raises(IndexError):
        ak._do.local_index(v2_array, 1)

    v2_array = ak.contents.unionarray.UnionArray(
        ak.index.Index(np.array([1, 1, 0, 0, 1, 0, 1], dtype=np.int8)),
        ak.index.Index(np.array([4, 3, 0, 1, 2, 2, 4, 100])),
        [
            ak.contents.recordarray.RecordArray(
                [ak.from_iter(["1", "2", "3"], highlevel=False)], ["nest"]
            ),
            ak.contents.recordarray.RecordArray(
                [
                    ak.contents.numpyarray.NumpyArray(
                        np.array([1.1, 2.2, 3.3, 4.4, 5.5])
                    )
                ],
                ["nest"],
            ),
        ],
    )

    assert to_list(ak._do.local_index(v2_array, 0)) == [0, 1, 2, 3, 4, 5, 6]
    assert (
        ak._do.local_index(v2_array.to_typetracer(), 0).form
        == ak._do.local_index(v2_array, 0).form
    )
    assert to_list(ak._do.local_index(v2_array, -1)) == [0, 1, 2, 3, 4, 5, 6]
    assert (
        ak._do.local_index(v2_array.to_typetracer(), -1).form
        == ak._do.local_index(v2_array, -1).form
    )

    with pytest.raises(IndexError):
        ak._do.local_index(v2_array, -2)
    with pytest.raises(IndexError):
        ak._do.local_index(v2_array, 1)



# BSD 3-Clause License; see https://github.com/scikit-hep/awkward-1.0/blob/main/LICENSE
from __future__ import annotations

import json
from collections.abc import Iterable
from itertools import permutations

import awkward as ak
import awkward._prettyprint
from awkward._behavior import find_record_typestr
from awkward._errors import deprecate
from awkward._parameters import parameters_are_equal, type_parameters_equal
from awkward._typing import Self, final
from awkward._util import UNSET
from awkward.types.type import Type


@final
class RecordType(Type):
    def copy(
        self,
        *,
        contents: list[Type] = UNSET,
        fields: list[str] | None = UNSET,
        parameters=UNSET,
        typestr=UNSET,
    ) -> Self:
        return RecordType(
            self._contents if contents is UNSET else contents,
            self._fields if fields is UNSET else fields,
            parameters=self._parameters if parameters is UNSET else parameters,
            typestr=self._typestr if typestr is UNSET else typestr,
        )

    def __init__(self, contents, fields, *, parameters=None, typestr=None):
        if not isinstance(contents, Iterable):
            raise TypeError(
                "{} 'contents' must be iterable, not {}".format(
                    type(self).__name__, repr(contents)
                )
            )
        if not isinstance(contents, list):
            contents = list(contents)
        for content in contents:
            if not isinstance(content, Type):
                raise TypeError(
                    "{} all 'contents' must be Type subclasses, not {}".format(
                        type(self).__name__, repr(content)
                    )
                )
        if fields is not None and not isinstance(fields, Iterable):
            raise TypeError(
                "{} 'fields' must be iterable, not {}".format(
                    type(self).__name__, repr(contents)
                )
            )
        if parameters is not None and not isinstance(parameters, dict):
            raise TypeError(
                "{} 'parameters' must be of type dict or None, not {}".format(
                    type(self).__name__, repr(parameters)
                )
            )
        if typestr is not None and not isinstance(typestr, str):
            raise TypeError(
                "{} 'typestr' must be of type string or None, not {}".format(
                    type(self).__name__, repr(typestr)
                )
            )
        self._contents = contents
        self._fields = fields
        self._parameters = parameters
        self._typestr = typestr

    @property
    def contents(self):
        return self._contents

    @property
    def fields(self):
        return self._fields

    @property
    def is_tuple(self):
        return self._fields is None

    _str_parameters_exclude = ("__categorical__", "__record__")

    def _str(self, indent, compact, behavior):
        if self._typestr is not None:
            deprecate("typestr argument is deprecated", "2.4.0")

        typestr = find_record_typestr(behavior, self._parameters, self._typestr)
        if typestr is not None:
            out = [typestr]

        else:
            if compact:
                pre, post = "", ""
            else:
                pre, post = "\n" + indent + "    ", "\n" + indent

            children = []
            for i, x in enumerate(self._contents):
                if i + 1 < len(self._contents):
                    if compact:
                        y = [*x._str(indent, compact, behavior), ", "]
                    else:
                        y = [
                            *x._str(indent + "    ", compact, behavior),
                            ",\n",
                            indent,
                            "    ",
                        ]
                else:
                    if compact:
                        y = x._str(indent, compact, behavior)
                    else:
                        y = x._str(indent + "    ", compact, behavior)
                children.append(y)

            params = self._str_parameters()
            name = self.parameter("__record__")

            if name is not None:
                if (
                    not ak._prettyprint.is_identifier.match(name)
                    or name
                    in (
                        "unknown",
                        "string",
                        "bytes",
                        "option",
                        "tuple",
                        "struct",
                        "union",
                        "categorical",
                    )
                    or name in ak.types.numpytype._primitive_to_dtype_dict
                ):
                    if params is None:
                        params = 'parameters={"__record__": ' + json.dumps(name) + "}"
                    else:
                        params = (
                            'parameters={"__record__": '
                            + json.dumps(name)
                            + ", "
                            + params[12:]
                        )
                    name = None

            if not self.is_tuple:
                pairs = []
                for k, v in zip(self._fields, children):
                    if ak._prettyprint.is_identifier.match(k) is None:
                        key_str = json.dumps(k)
                    else:
                        key_str = k
                    pairs.append([key_str, ": ", *v])
                flat_pairs = [y for x in pairs for y in x]

            if params is None:
                if self.is_tuple:
                    flat_children = [y for x in children for y in x]
                    if name is None:
                        out = ["(", pre, *flat_children, post, ")"]
                    else:
                        out = [name, "[", pre, *flat_children, post, "]"]
                else:
                    if name is None:
                        out = ["{", pre, *flat_pairs, post, "}"]
                    else:
                        out = [name, "[", pre, *flat_pairs, post, "]"]

            else:
                if self.is_tuple:
                    flat_children = [y for x in children for y in x]
                    if name is None:
                        out = ["tuple[[", pre, *flat_children, post, "], ", params, "]"]
                    else:
                        c = "" if len(self._contents) == 0 else ", "
                        out = [name, "[", pre, *flat_children, c, post, params, "]"]
                else:
                    if name is None:
                        out = ["struct[{", pre, *flat_pairs, post, "}, ", params, "]"]
                    else:
                        c = "" if len(self._contents) == 0 else ", "
                        out = [name, "[", pre, *flat_pairs, c, post, params, "]"]

        return [self._str_categorical_begin(), *out, self._str_categorical_end()]

    def __repr__(self):
        args = [repr(self._contents), repr(self._fields), *self._repr_args()]
        return "{}({})".format(type(self).__name__, ", ".join(args))

    def _is_equal_to(self, other, all_parameters: bool) -> bool:
        if not isinstance(other, type(self)):
            return False

        compare_parameters = (
            parameters_are_equal if all_parameters else type_parameters_equal
        )
        if not compare_parameters(self._parameters, other._parameters):
            return False

        # Both tuples
        if self.is_tuple and other.is_tuple:
            return all(
                this._is_equal_to(that, all_parameters)
                for this, that in zip(self._contents, other._contents)
            )
        # Both records
        elif not (self.is_tuple or other.is_tuple):
            if set(self._fields) != set(other._fields):
                return False

            self_contents = [self.content(f) for f in self._fields]
            other_contents = [other.content(f) for f in other._fields]

            return any(
                all(
                    this._is_equal_to(that, all_parameters)
                    for this, that in zip(self_contents, contents)
                )
                for contents in permutations(other_contents)
            )
        # Mixed
        else:
            return False

    def index_to_field(self, index):
        return ak.forms.RecordForm.index_to_field(self, index)

    def field_to_index(self, field):
        return ak.forms.RecordForm.field_to_index(self, field)

    def has_field(self, field):
        return ak.forms.RecordForm.has_field(self, field)

    def content(self, index_or_field):
        return ak.forms.RecordForm.content(self, index_or_field)



# BSD 3-Clause License; see https://github.com/scikit-hep/awkward-1.0/blob/main/LICENSE

import numpy as np  # noqa: F401
import pytest  # noqa: F401

import awkward as ak


def test():
    x = ak.highlevel.Array([1, 2, 3, None, 4])
    assert ak.operations.argmax(x) == 4



# BSD 3-Clause License; see https://github.com/scikit-hep/awkward-1.0/blob/main/LICENSE

import pytest  # noqa: F401

import awkward as ak


def test_axis_none():
    record = ak.zip({"x": [1, None], "y": [2, 3]})
    assert ak.fill_none(record, 0, axis=None).to_list() == [
        {"x": 1, "y": 2},
        {"x": 0, "y": 3},
    ]


def test_axis_last():
    record = ak.zip({"x": [1, None], "y": [2, 3]})
    assert ak.fill_none(record, 0, axis=-1).to_list() == [
        {"x": 1, "y": 2},
        {"x": 0, "y": 3},
    ]


def test_option_outside_record():
    record = ak.zip({"x": [1, 4], "y": [2, 3]}).mask[[True, False]]
    assert ak.fill_none(record, 0, axis=-1).to_list() == [{"x": 1, "y": 2}, 0]



# BSD 3-Clause License; see https://github.com/scikit-hep/awkward-1.0/blob/main/LICENSE

import pytest  # noqa: F401

import awkward as ak


def test():
    result = ak.broadcast_arrays()
    assert result == []



# BSD 3-Clause License; see https://github.com/scikit-hep/awkward-1.0/blob/main/LICENSE

import pytest

import awkward as ak
from awkward.types.listtype import ListType
from awkward.types.numpytype import NumpyType
from awkward.types.optiontype import OptionType
from awkward.types.recordtype import RecordType
from awkward.types.regulartype import RegularType
from awkward.types.uniontype import UnionType
from awkward.types.unknowntype import UnknownType


def test_primitive_1():
    text = "int64"
    parsedtype = ak.types.from_datashape(text, highlevel=False)
    assert isinstance(parsedtype, ak.types.NumpyType)
    assert (str(parsedtype)) == text


def test_primitive_2():
    text = 'int64[parameters={"wonky": ["parameter", 3.14]}]'
    parsedtype = ak.types.from_datashape(text, highlevel=False)
    assert isinstance(parsedtype, ak.types.NumpyType)
    assert (str(parsedtype)) == text


def test_unknown_1():
    text = "unknown"
    parsedtype = ak.types.from_datashape(text, highlevel=False)
    assert isinstance(parsedtype, ak.types.UnknownType)
    assert (str(parsedtype)) == text


def test_unknown_2():
    text = 'unknown[parameters={"wonky": ["parameter", 3.14]}]'
    with pytest.raises(TypeError):
        ak.types.from_datashape(text, highlevel=False)


def test_record_tuple_1():
    text = "(int64)"
    parsedtype = ak.types.from_datashape(text, highlevel=False)
    assert isinstance(parsedtype, ak.types.RecordType)
    assert str(parsedtype) == text


def test_record_tuple_2():
    text = '(int64[parameters={"wonky": ["bla", 1, 2]}])'
    parsedtype = ak.types.from_datashape(text, highlevel=False)
    assert isinstance(parsedtype, ak.types.RecordType)
    assert str(parsedtype) == text


def test_record_tuple_3():
    text = '(int64, int64[parameters={"wonky": ["bla", 1, 2]}])'
    parsedtype = ak.types.from_datashape(text, highlevel=False)
    assert isinstance(parsedtype, ak.types.RecordType)
    assert str(parsedtype) == text


def test_record_dict_1():
    text = '{"1": int64}'
    parsedtype = ak.types.from_datashape(text, highlevel=False)
    assert isinstance(parsedtype, ak.types.RecordType)
    assert str(parsedtype) == text


def test_record_dict_2():
    text = '{bla: int64[parameters={"wonky": ["bla", 1, 2]}]}'
    parsedtype = ak.types.from_datashape(text, highlevel=False)
    assert isinstance(parsedtype, ak.types.RecordType)
    assert str(parsedtype) == text


def test_record_dict_3():
    text = '{bla: int64[parameters={"wonky": ["bla", 1, 2]}], foo: int64}'
    parsedtype = ak.types.from_datashape(text, highlevel=False)
    assert isinstance(parsedtype, ak.types.RecordType)
    assert str(parsedtype) == text


def test_record_parmtuple_1():
    text = 'tuple[[int64[parameters={"xkcd": [11, 12, 13]}]], parameters={"wonky": ["bla", 1, 2]}]'
    parsedtype = ak.types.from_datashape(text, highlevel=False)
    assert isinstance(parsedtype, ak.types.RecordType)
    assert str(parsedtype) == text


def test_record_parmtuple_2():
    text = 'tuple[[int64, int64], parameters={"wonky": ["bla", 1, 2]}]'
    parsedtype = ak.types.from_datashape(text, highlevel=False)
    assert isinstance(parsedtype, ak.types.RecordType)
    assert str(parsedtype) == text


def test_record_struct_1():
    text = 'struct[{"1": int64[parameters={"xkcd": [11, 12, 13]}]}, parameters={"wonky": ["bla", 1, 2]}]'
    parsedtype = ak.types.from_datashape(text, highlevel=False)
    assert isinstance(parsedtype, ak.types.RecordType)
    assert str(parsedtype) == text


def test_record_struct_2():
    text = 'struct[{"1": int64[parameters={"xkcd": [11, 12, 13]}], "2": int64}, parameters={"wonky": ["bla", 1, 2]}]'
    parsedtype = ak.types.from_datashape(text, highlevel=False)
    assert isinstance(parsedtype, ak.types.RecordType)
    assert str(parsedtype) == text


def test_option_numpy_1():
    text = "?int64"
    parsedtype = ak.types.from_datashape(text, highlevel=False)
    assert isinstance(parsedtype, ak.types.OptionType)
    assert str(parsedtype) == text


def test_option_numpy_2():
    text = '?int64[parameters={"wonky": [1, 2, 3]}]'
    parsedtype = ak.types.from_datashape(text, highlevel=False)
    assert isinstance(parsedtype, ak.types.OptionType)
    assert str(parsedtype) == text


def test_option_numpy_1_parm():
    text = 'option[int64, parameters={"foo": "bar"}]'
    parsedtype = ak.types.from_datashape(text, highlevel=False)
    assert isinstance(parsedtype, ak.types.OptionType)
    assert str(parsedtype) == text


def test_option_numpy_2_parm():
    text = 'option[int64[parameters={"wonky": [1, 2]}], parameters={"foo": "bar"}]'
    parsedtype = ak.types.from_datashape(text, highlevel=False)
    assert isinstance(parsedtype, ak.types.OptionType)
    assert str(parsedtype) == text


def test_option_unknown_1():
    text = "?unknown"
    parsedtype = ak.types.from_datashape(text, highlevel=False)
    assert isinstance(parsedtype, ak.types.OptionType)
    assert str(parsedtype) == text


def test_option_unknown_2():
    text = '?unknown[parameters={"foo": "bar"}]'
    with pytest.raises(TypeError):
        ak.types.from_datashape(text, highlevel=False)


def test_option_unknown_1_parm():
    text = 'option[unknown, parameters={"foo": "bar"}]'
    parsedtype = ak.types.from_datashape(text, highlevel=False)
    assert isinstance(parsedtype, ak.types.OptionType)
    assert str(parsedtype) == text


def test_option_unknown_2_parm():
    text = 'option[unknown, parameters={"foo": "bar"}]'
    parsedtype = ak.types.from_datashape(text, highlevel=False)
    assert isinstance(parsedtype, ak.types.OptionType)
    assert str(parsedtype) == text


def test_regular_numpy_1():
    text = "5 * int64"
    parsedtype = ak.types.from_datashape(text, highlevel=False)
    assert isinstance(parsedtype, ak.types.RegularType)
    assert str(parsedtype) == text


def test_regular_numpy_2():
    text = '5 * int64[parameters={"bar": "foo"}]'
    parsedtype = ak.types.from_datashape(text, highlevel=False)
    assert isinstance(parsedtype, ak.types.RegularType)
    assert str(parsedtype) == text


def test_regular_numpy_2_parm():
    text = '[0 * int64[parameters={"foo": "bar"}], parameters={"bla": "bloop"}]'
    parsedtype = ak.types.from_datashape(text, highlevel=False)
    assert isinstance(parsedtype, ak.types.RegularType)
    assert str(parsedtype) == text


def test_regular_unknown_1_parm():
    text = '[0 * unknown, parameters={"foo": "bar"}]'
    parsedtype = ak.types.from_datashape(text, highlevel=False)
    assert isinstance(parsedtype, ak.types.RegularType)
    assert str(parsedtype) == text


def test_list_numpy_1():
    text = "var * float64"
    parsedtype = ak.types.from_datashape(text, highlevel=False)
    assert isinstance(parsedtype, ak.types.ListType)
    assert str(parsedtype) == text


def test_list_numpy_1_parm():
    text = '[var * float64[parameters={"wonky": "boop"}], parameters={"foo": "bar"}]'
    parsedtype = ak.types.from_datashape(text, highlevel=False)
    assert isinstance(parsedtype, ak.types.ListType)
    assert str(parsedtype) == text


def test_union_numpy_empty_1():
    text = 'union[float64[parameters={"wonky": "boop"}], unknown]'
    parsedtype = ak.types.from_datashape(text, highlevel=False)
    assert isinstance(parsedtype, ak.types.UnionType)
    assert str(parsedtype) == text


def test_union_numpy_empty_1_parm():
    text = 'union[float64[parameters={"wonky": "boop"}], unknown, parameters={"pratyush": "das"}]'
    parsedtype = ak.types.from_datashape(text, highlevel=False)
    assert isinstance(parsedtype, ak.types.UnionType)
    assert str(parsedtype) == text


def test_arraytype_string():
    text = str(ak.Array([["one", "two", "three"], [], ["four", "five"]]).type)
    parsedtype = ak.types.from_datashape(text, highlevel=True)
    assert isinstance(parsedtype, ak.types.ArrayType)
    assert str(parsedtype) == text


def test_arraytype_bytestring():
    text = str(ak.Array([[b"one", b"two", b"three"], [], [b"four", b"five"]]).type)
    parsedtype = ak.types.from_datashape(text, highlevel=True)
    assert isinstance(parsedtype, ak.types.ArrayType)
    assert str(parsedtype) == text


def test_arraytype_categorical_1():
    text = str(
        ak.operations.ak_to_categorical.to_categorical(
            ak.Array(["one", "one", "two", "three", "one", "three"])
        ).type
    )
    parsedtype = ak.types.from_datashape(text, highlevel=True)
    assert isinstance(parsedtype, ak.types.ArrayType)
    assert str(parsedtype) == text


def test_arraytype_categorical_2():
    text = str(
        ak.operations.ak_to_categorical.to_categorical(
            ak.Array([1.1, 1.1, 2.2, 3.3, 1.1, 3.3])
        ).type
    )
    parsedtype = ak.types.from_datashape(text, highlevel=True)
    assert isinstance(parsedtype, ak.types.ArrayType)
    assert str(parsedtype) == text


def test_arraytype_record_1():
    text = str(
        ak.Array(
            [{"x": 1, "y": 1.1}, {"x": 2, "y": 2.2}, {"x": 3, "y": 3.3}],
            with_name="Thingy",
        ).type
    )
    parsedtype = ak.types.from_datashape(text, highlevel=True)
    assert isinstance(parsedtype, ak.types.ArrayType)
    assert str(parsedtype) == text


def test_arraytype_record_2():
    text = str(
        ak.Array(
            [[{"x": 1, "y": 1.1}, {"x": 2, "y": 2.2}], [], [{"x": 3, "y": 3.3}]],
            with_name="Thingy",
        ).type
    )
    parsedtype = ak.types.from_datashape(text, highlevel=True)
    assert isinstance(parsedtype, ak.types.ArrayType)
    assert str(parsedtype) == text


def test_arraytype_1():
    text = str(ak.Array([[1, 2, 3], None, [4, 5]]).type)
    parsedtype = ak.types.from_datashape(text, highlevel=True)
    assert isinstance(parsedtype, ak.types.ArrayType)
    assert str(parsedtype) == text


def test_arraytype_2():
    text = str(
        ak.with_parameter(ak.Array([[1, 2, 3], [], [4, 5]]), "wonky", "string").type
    )
    parsedtype = ak.types.from_datashape(text, highlevel=True)
    assert isinstance(parsedtype, ak.types.ArrayType)
    assert str(parsedtype) == text


def test_arraytype_3():
    text = str(
        ak.with_parameter(
            ak.Array([[1, 2, 3], [], [4, 5]]), "wonky", {"other": "JSON"}
        ).type
    )
    parsedtype = ak.types.from_datashape(text, highlevel=False)
    assert str(parsedtype) == text


def test_arraytype_4():
    text = str(
        ak.with_parameter(ak.Array([[1, 2, 3], None, [4, 5]]), "wonky", "string").type
    )
    parsedtype = ak.types.from_datashape(text, highlevel=False)
    assert str(parsedtype) == text


def test_arraytype_5():
    text = str(
        ak.with_parameter(ak.Array([1, 2, 3, None, 4, 5]), "wonky", "string").type
    )
    parsedtype = ak.types.from_datashape(text, highlevel=False)
    assert str(parsedtype) == text


def test_arraytype_6():
    text = str(ak.with_parameter(ak.Array([1, 2, 3, 4, 5]), "wonky", "string").type)
    parsedtype = ak.types.from_datashape(text, highlevel=False)
    assert str(parsedtype) == text


def test_arraytype_7():
    text = str(ak.Array([1, 2, 3, None, 4, 5]).type)
    parsedtype = ak.types.from_datashape(text, highlevel=False)
    assert str(parsedtype) == text


def test_arraytype_8():
    text = str(
        ak.with_parameter(
            ak.Array([{"x": 1, "y": 1.1}, {"x": 2, "y": 2.2}, {"x": 3, "y": 3.3}]),
            "wonky",
            "string",
        ).type
    )
    parsedtype = ak.types.from_datashape(text, highlevel=False)
    assert str(parsedtype) == text


def test_arraytype_9():
    text = str(ak.Array([(1, 1.1), (2, 2.2), (3, 3.3)]).type)
    parsedtype = ak.types.from_datashape(text, highlevel=False)
    assert str(parsedtype) == text


def test_arraytype_10():
    text = str(
        ak.with_parameter(
            ak.Array([(1, 1.1), (2, 2.2), (3, 3.3)]), "wonky", "string"
        ).type
    )
    parsedtype = ak.types.from_datashape(text, highlevel=False)
    assert str(parsedtype) == text


def test_arraytype_11():
    text = str(ak.Array([[(1, 1.1), (2, 2.2)], [], [(3, 3.3)]]).type)
    parsedtype = ak.types.from_datashape(text, highlevel=False)
    assert str(parsedtype) == text


def test_arraytype_12():
    text = str(ak.to_regular(ak.Array([[1, 2], [3, 4], [5, 6]])).type)
    parsedtype = ak.types.from_datashape(text, highlevel=False)
    assert str(parsedtype) == text


def test_arraytype_13():
    text = str(
        ak.with_parameter(
            ak.to_regular(ak.Array([[1, 2], [3, 4], [5, 6]])), "wonky", "string"
        ).type
    )
    parsedtype = ak.types.from_datashape(text, highlevel=False)
    assert str(parsedtype) == text


def test_arraytype_14():
    text = str(
        ak.with_parameter(
            ak.Array([1, 2, 3, [1], [1, 2], [1, 2, 3]]), "wonky", "string"
        ).type
    )
    parsedtype = ak.types.from_datashape(text, highlevel=False)
    assert str(parsedtype) == text


def test_arraytype_15():
    text = str(
        ak.with_parameter(
            ak.Array([1, 2, 3, None, [1], [1, 2], [1, 2, 3]]), "wonky", "string"
        ).type
    )
    parsedtype = ak.types.from_datashape(text, highlevel=False)
    assert str(parsedtype) == text


def test_arraytype_16():
    text = str(ak.Array([1, 2, 3, None, [1], [1, 2], [1, 2, 3]]).type)
    parsedtype = ak.types.from_datashape(text, highlevel=False)
    assert str(parsedtype) == text


def test_arraytype_17():
    text = str(ak.Array([1, 2, 3, None, [], [], []]).type)
    parsedtype = ak.types.from_datashape(text, highlevel=False)
    assert str(parsedtype) == text


def test_string():
    text = "string"
    parsedtype = ak.types.from_datashape(text, highlevel=False)
    assert isinstance(parsedtype, ak.types.ListType)
    assert str(parsedtype) == text


def test_hardcoded():
    text = "var * string"
    parsedtype = ak.types.from_datashape(text, highlevel=False)
    assert isinstance(parsedtype, ak.types.ListType)
    assert str(parsedtype) == text


def test_record_highlevel():
    text = "Thingy[x: int64, y: float64]"
    parsedtype = ak.types.from_datashape(text, highlevel=True)
    assert isinstance(parsedtype, ak.types.ScalarType)
    assert str(parsedtype) == text


def test_numpytype_int32():
    t = NumpyType("int32")
    assert str(ak.types.from_datashape(str(t), highlevel=False)) == str(t)


def test_numpytype_datetime64():
    t = NumpyType("datetime64")
    assert str(ak.types.from_datashape(str(t), highlevel=False)) == str(t)


def test_numpytype_datetime64_10s():
    t = NumpyType("datetime64[10s]")
    assert str(ak.types.from_datashape(str(t), highlevel=False)) == str(t)


def test_numpytype_int32_parameter():
    t = NumpyType("int32", parameters={"__array__": "Something"})
    assert str(ak.types.from_datashape(str(t), highlevel=False)) == str(t)


def test_numpytype_datetime64_parameter():
    t = NumpyType("datetime64", parameters={"__array__": "Something"})
    assert str(ak.types.from_datashape(str(t), highlevel=False)) == str(t)


def test_numpytype_datetime64_10s_parameter():
    t = NumpyType("datetime64[10s]", parameters={"__array__": "Something"})
    assert str(ak.types.from_datashape(str(t), highlevel=False)) == str(t)


def test_numpytype_int32_categorical():
    t = NumpyType("int32", parameters={"__categorical__": True})
    assert str(ak.types.from_datashape(str(t), highlevel=False)) == str(t)


def test_numpytype_int32_parameters_categorical():
    t = NumpyType(
        "int32", parameters={"__array__": "Something", "__categorical__": True}
    )
    assert str(ak.types.from_datashape(str(t), highlevel=False)) == str(t)


def test_unknowntype():
    t = UnknownType()
    assert str(ak.types.from_datashape(str(t), highlevel=False)) == str(t)


def test_unknowntype_parameter():
    with pytest.raises(TypeError):
        UnknownType(parameters={"__array__": "Something"})


def test_unknowntype_categorical():
    with pytest.raises(TypeError):
        UnknownType(parameters={"__categorical__": True})


def test_unknowntype_categorical_parameter():
    with pytest.raises(TypeError):
        UnknownType(parameters={"__array__": "Something", "__categorical__": True})


def test_regulartype_numpytype():
    t = RegularType(NumpyType("int32"), 5)
    assert str(ak.types.from_datashape(str(t), highlevel=False)) == str(t)


def test_regulartype_numpytype_parameter():
    t = RegularType(NumpyType("int32"), 5, parameters={"__array__": "Something"})
    assert str(ak.types.from_datashape(str(t), highlevel=False)) == str(t)


def test_regulartype_numpytype_categorical():
    t = RegularType(NumpyType("int32"), 5, parameters={"__categorical__": True})
    assert str(ak.types.from_datashape(str(t), highlevel=False)) == str(t)


def test_regulartype_numpytype_categorical_parameter():
    t = RegularType(
        NumpyType("int32"),
        5,
        parameters={"__categorical__": True, "__array__": "Something"},
    )
    assert str(ak.types.from_datashape(str(t), highlevel=False)) == str(t)


def test_listtype_numpytype():
    t = ListType(NumpyType("int32"))
    assert str(ak.types.from_datashape(str(t), highlevel=False)) == str(t)


def test_listtype_numpytype_parameter():
    t = ListType(NumpyType("int32"), parameters={"__array__": "Something"})
    assert str(ak.types.from_datashape(str(t), highlevel=False)) == str(t)


def test_listtype_numpytype_categorical():
    t = ListType(NumpyType("int32"), parameters={"__categorical__": True})
    assert str(ak.types.from_datashape(str(t), highlevel=False)) == str(t)


def test_listtype_numpytype_categorical_parameter():
    t = ListType(
        NumpyType("int32"),
        parameters={"__categorical__": True, "__array__": "Something"},
    )
    assert str(ak.types.from_datashape(str(t), highlevel=False)) == str(t)


def test_varlen_string():
    t = ListType(
        NumpyType("uint8", parameters={"__array__": "char"}),
        parameters={"__array__": "string"},
    )
    assert str(ak.types.from_datashape(str(t), highlevel=False)) == str(t)


def test_varlen_bytestring():
    t = ListType(
        NumpyType("uint8", parameters={"__array__": "char"}),
        parameters={"__array__": "bytestring"},
    )
    assert str(ak.types.from_datashape(str(t), highlevel=False)) == str(t)


def test_fixedlen_string():
    t = RegularType(
        NumpyType("uint8", parameters={"__array__": "char"}),
        5,
        parameters={"__array__": "string"},
    )
    assert str(ak.types.from_datashape(str(t), highlevel=False)) == str(t)


def test_fixedlen_bytestring():
    t = RegularType(
        NumpyType("uint8", parameters={"__array__": "byte"}),
        5,
        parameters={"__array__": "bytestring"},
    )
    assert str(ak.types.from_datashape(str(t), highlevel=False)) == str(t)


def test_char():
    t = NumpyType("uint8", parameters={"__array__": "char"})
    assert str(ak.types.from_datashape(str(t), highlevel=False)) == str(t)


def test_byte():
    t = NumpyType("uint8", parameters={"__array__": "byte"})
    assert str(ak.types.from_datashape(str(t), highlevel=False)) == str(t)


def test_optiontype_numpytype_int32():
    t = OptionType(NumpyType("int32"))
    assert str(ak.types.from_datashape(str(t), highlevel=False)) == str(t)


def test_optiontype_numpytype_int32_parameters():
    t = OptionType(NumpyType("int32"), parameters={"__array__": "Something"})
    assert str(ak.types.from_datashape(str(t), highlevel=False)) == str(t)


def test_optiontype_numpytype_int32_categorical():
    t = OptionType(NumpyType("int32"), parameters={"__categorical__": True})
    assert str(ak.types.from_datashape(str(t), highlevel=False)) == str(t)


def test_optiontype_numpytype_int32_categorical_parameters():
    t = OptionType(
        NumpyType("int32"),
        parameters={"__array__": "Something", "__categorical__": True},
    )
    assert str(ak.types.from_datashape(str(t), highlevel=False)) == str(t)


def test_option_varlen_string():
    t = OptionType(
        ListType(
            NumpyType("uint8", parameters={"__array__": "char"}),
            parameters={"__array__": "string"},
        )
    )
    assert str(ak.types.from_datashape(str(t), highlevel=False)) == str(t)


def test_option_varlen_string_parameters():
    t = OptionType(
        ListType(
            NumpyType("uint8", parameters={"__array__": "char"}),
            parameters={"__array__": "string"},
        ),
        parameters={"__array__": "Something"},
    )
    assert str(ak.types.from_datashape(str(t), highlevel=False)) == str(t)


def test_record_empty():
    t = RecordType([], None)
    assert str(ak.types.from_datashape(str(t), highlevel=False)) == str(t)


def test_record_fields_empty():
    t = RecordType([], [])
    assert str(ak.types.from_datashape(str(t), highlevel=False)) == str(t)


def test_record_int32():
    t = RecordType([NumpyType("int32")], None)
    assert str(ak.types.from_datashape(str(t), highlevel=False)) == str(t)


def test_record_int32_float64():
    t = RecordType([NumpyType("int32"), NumpyType("float64")], None)
    assert str(ak.types.from_datashape(str(t), highlevel=False)) == str(t)


def test_record_fields_int32():
    t = RecordType([NumpyType("int32")], ["one"])
    assert str(ak.types.from_datashape(str(t), highlevel=False)) == str(t)


def test_record_fields_int32_float64():
    t = RecordType([NumpyType("int32"), NumpyType("float64")], ["one", "t w o"])
    assert str(ak.types.from_datashape(str(t), highlevel=False)) == str(t)


def test_record_empty_parameters():
    t = RecordType([], None, parameters={"p": [123]})
    assert str(ak.types.from_datashape(str(t), highlevel=False)) == str(t)


def test_record_fields_empty_parameters():
    t = RecordType([], [], parameters={"p": [123]})
    assert str(ak.types.from_datashape(str(t), highlevel=False)) == str(t)


def test_record_int32_parameters():
    t = RecordType([NumpyType("int32")], None, parameters={"p": [123]})
    assert str(ak.types.from_datashape(str(t), highlevel=False)) == str(t)


def test_record_int32_float64_parameters():
    t = RecordType(
        [NumpyType("int32"), NumpyType("float64")], None, parameters={"p": [123]}
    )
    assert str(ak.types.from_datashape(str(t), highlevel=False)) == str(t)


def test_record_fields_int32_parameters():
    t = RecordType([NumpyType("int32")], ["one"], parameters={"p": [123]})
    assert str(ak.types.from_datashape(str(t), highlevel=False)) == str(t)


def test_record_fields_int32_float64_parameters():
    t = RecordType(
        [NumpyType("int32"), NumpyType("float64")],
        ["one", "t w o"],
        parameters={"p": [123]},
    )
    assert str(ak.types.from_datashape(str(t), highlevel=False)) == str(t)


def test_named_record_empty():
    t = RecordType([], None, parameters={"__record__": "Name"})
    assert str(ak.types.from_datashape(str(t), highlevel=False)) == str(t)


def test_named_record_int32():
    t = RecordType([NumpyType("int32")], None, parameters={"__record__": "Name"})
    assert str(ak.types.from_datashape(str(t), highlevel=False)) == str(t)


def test_named_record_int32_float64():
    t = RecordType(
        [NumpyType("int32"), NumpyType("float64")],
        None,
        parameters={"__record__": "Name"},
    )
    assert str(ak.types.from_datashape(str(t), highlevel=False)) == str(t)


def test_named_record_fields_int32():
    t = RecordType([NumpyType("int32")], ["one"], parameters={"__record__": "Name"})
    assert str(ak.types.from_datashape(str(t), highlevel=False)) == str(t)


def test_named_record_fields_int32_float64():
    t = RecordType(
        [NumpyType("int32"), NumpyType("float64")],
        ["one", "t w o"],
        parameters={"__record__": "Name"},
    )
    assert str(ak.types.from_datashape(str(t), highlevel=False)) == str(t)


def test_named_record_empty_parameters():
    t = RecordType([], None, parameters={"__record__": "Name", "p": [123]})
    assert str(ak.types.from_datashape(str(t), highlevel=False)) == str(t)


def test_named_record_int32_parameters():
    t = RecordType(
        [NumpyType("int32")], None, parameters={"__record__": "Name", "p": [123]}
    )
    assert str(ak.types.from_datashape(str(t), highlevel=False)) == str(t)


def test_named_record_int32_float64_parameters():
    t = RecordType(
        [NumpyType("int32"), NumpyType("float64")],
        None,
        parameters={"__record__": "Name", "p": [123]},
    )
    assert str(ak.types.from_datashape(str(t), highlevel=False)) == str(t)


def test_named_record_fields_int32_parameters():
    t = RecordType(
        [NumpyType("int32")], ["one"], parameters={"__record__": "Name", "p": [123]}
    )
    assert str(ak.types.from_datashape(str(t), highlevel=False)) == str(t)


def test_named_record_fields_int32_float64_parameters():
    t = RecordType(
        [NumpyType("int32"), NumpyType("float64")],
        ["one", "t w o"],
        parameters={"__record__": "Name", "p": [123]},
    )
    assert str(ak.types.from_datashape(str(t), highlevel=False)) == str(t)


def test_union_empty():
    t = UnionType([])
    assert str(ak.types.from_datashape(str(t), highlevel=False)) == str(t)


def test_union_float64():
    t = UnionType([NumpyType("float64")])
    assert str(ak.types.from_datashape(str(t), highlevel=False)) == str(t)


def test_union_float64_datetime64():
    t = UnionType(
        [NumpyType("float64"), NumpyType("datetime64")],
    )
    assert str(ak.types.from_datashape(str(t), highlevel=False)) == str(t)


def test_union_float64_parameters():
    t = UnionType([NumpyType("float64")], parameters={"__array__": "Something"})
    assert str(ak.types.from_datashape(str(t), highlevel=False)) == str(t)


def test_union_float64_datetime64_parameters():
    t = UnionType(
        [NumpyType("float64"), NumpyType("datetime64")],
        parameters={"__array__": "Something"},
    )
    assert str(ak.types.from_datashape(str(t), highlevel=False)) == str(t)



# BSD 3-Clause License; see https://github.com/scikit-hep/awkward-1.0/blob/main/LICENSE

import numpy as np  # noqa: F401
import pytest  # noqa: F401

import awkward as ak

to_list = ak.operations.to_list


def test():
    assert to_list(
        ak.operations.prod(
            ak.highlevel.Array([[[2, 3, 5]], [[7], [11]], [[]]]), axis=-1
        )
    ) == [
        [30],
        [7, 11],
        [1],
    ]

    assert to_list(
        ak.operations.prod(ak.highlevel.Array([[[2, 3, 5]], [[7], [11]], []]), axis=-1)
    ) == [
        [30],
        [7, 11],
        [],
    ]



# BSD 3-Clause License; see https://github.com/scikit-hep/awkward-1.0/blob/main/LICENSE
from __future__ import annotations

from awkward._behavior import find_array_typestr
from awkward._errors import deprecate
from awkward._parameters import (
    parameters_are_equal,
    parameters_union,
    type_parameters_equal,
)
from awkward._typing import Self, final
from awkward._util import UNSET
from awkward.types.listtype import ListType
from awkward.types.regulartype import RegularType
from awkward.types.type import Type
from awkward.types.uniontype import UnionType


@final
class OptionType(Type):
    def copy(self, *, content: Type = UNSET, parameters=UNSET, typestr=UNSET) -> Self:
        return OptionType(
            self._content if content is UNSET else content,
            parameters=self._parameters if parameters is UNSET else parameters,
            typestr=self._typestr if typestr is UNSET else typestr,
        )

    def __init__(self, content, *, parameters=None, typestr=None):
        if not isinstance(content, Type):
            raise TypeError(
                "{} 'content' must be a Type subclass, not {}".format(
                    type(self).__name__, repr(content)
                )
            )
        if parameters is not None and not isinstance(parameters, dict):
            raise TypeError(
                "{} 'parameters' must be of type dict or None, not {}".format(
                    type(self).__name__, repr(parameters)
                )
            )
        if typestr is not None and not isinstance(typestr, str):
            raise TypeError(
                "{} 'typestr' must be of type string or None, not {}".format(
                    type(self).__name__, repr(typestr)
                )
            )
        self._content = content
        self._parameters = parameters
        self._typestr = typestr

    @property
    def content(self):
        return self._content

    def _str(self, indent, compact, behavior):
        if self._typestr is not None:
            deprecate("typestr argument is deprecated", "2.4.0")

        typestr = find_array_typestr(behavior, self._parameters, self._typestr)

        head = []
        tail = []
        if typestr is not None:
            content_out = [typestr]

        else:
            content_out = self._content._str(indent, compact, behavior)
            params = self._str_parameters()
            if params is None:
                if isinstance(
                    self._content, (RegularType, ListType)
                ) and self._content.parameter("__array__") not in (
                    "string",
                    "bytestring",
                    "char",
                    "byte",
                ):
                    head = ["option["]
                    tail = ["]"]
                else:
                    head = ["?"]

            else:
                head = ["option["]
                tail = [f", {params}]"]

        return [
            *head,
            self._str_categorical_begin(),
            *content_out,
            self._str_categorical_end(),
            *tail,
        ]

    def __repr__(self):
        args = [repr(self._content), *self._repr_args()]
        return "{}({})".format(type(self).__name__, ", ".join(args))

    def simplify_option_union(self):
        if isinstance(self._content, UnionType):
            contents = []
            for content in self._content.contents:
                if isinstance(content, OptionType):
                    typestr = (
                        content._typestr if self._typestr is None else self._typestr
                    )
                    contents.append(
                        OptionType(
                            content.content,
                            parameters=parameters_union(
                                self._parameters, content._parameters
                            ),
                            typestr=typestr,
                        )
                    )

                else:
                    contents.append(
                        OptionType(
                            content, parameters=self._parameters, typestr=self._typestr
                        )
                    )

            return UnionType(
                contents,
                parameters=self._content.parameters,
                typestr=self._content.typestr,
            )

        else:
            return self

    def _is_equal_to(self, other, all_parameters: bool) -> bool:
        compare_parameters = (
            parameters_are_equal if all_parameters else type_parameters_equal
        )
        return (
            isinstance(other, type(self))
            and compare_parameters(self._parameters, other._parameters)
            and self._content._is_equal_to(
                other._content, all_parameters=all_parameters
            )
        )
